function [model, w, SN, zN] = LMProMPs_init(datasets, hyperparameters, DoF, NumGaussians)
%function [model, w, SN, zN] = LMProMPs_init(datasets, maxDimLatent, hyperparameters)
%
% This function computes an initial model. 
% For each demonstration the regression vector w, its covariance SN and a
% class assignment label zN is computed. 
%
% In the model is composed of an offset and a linear term: w = wo + wl h
%
%**************************************************************************

    K = datasets(1).NumCluster;     %(initial) number of clusters
    L = size(datasets,2);           %number of demonstrations in total

    M = size(datasets(1).PSI_X,2);  % must be equal for all tasks: DOF * NumberOfGaussians

    model = hyperparameters;

    model.K = K;
    model.L = L;
    model.M = M;
    model.DoF = DoF;
    model.NumGaussians = NumGaussians;
    
    flgEstHyperFromData = hyperparameters.flgEstHyperFromData;
    flgUseKMeans = hyperparameters.flgUseKMeans;
    
    [model, w, SN, zN] = get_approx_parameter(datasets,M,K,L, hyperparameters.maxDimLatent, ...
        model, flgEstHyperFromData, flgUseKMeans);
    
 end

function [mk,w,SN,zN] = get_approx_parameter(datasets,M,K,L,V, mk,use_approx_hypers, flgUseKMeans)

    mk.comp = struct([]);    
    
    % beta -> cannot be used since we consider only one trajectory per task
    if use_approx_hypers
        betas = zeros(1,L);
        for i=1:L 
            precision = var(datasets(i).Y,0,2);
            betas(i) = mean(precision);
            if isinf(betas(i)) || isnan(betas(i))
                betas(i) = 10 * min(precision);
            end
        end
        mk.beta = mean(betas)^-1;    
        if isinf(mk.beta) || isnan(mk.beta)
           mk.beta = 10^10; 
        end
    end
    
     % w
    w = zeros(M,L);
    yTrainInit = struct([]);
    XTrainInit = struct([]);
    
    centers = zeros(K, M); 
    indices = zeros(L,1);
    for i = 1:L
        X_train = datasets(i).PSI_X;
        y_train = datasets(i).Y;
        [w(:,i),~] = lin_reg_estimate_parameter(X_train,y_train);
        
        yTrainInit(i).data = y_train;
        XTrainInit(i).data = X_train;
        indices(i,1) = datasets(i).clusterId;
        centers(datasets(i).clusterId,:) = centers(datasets(i).clusterId,:) + w(:,i)';
    end
    for k=1:K
        centers(k,:) = centers(k,:) ./ sum(indices==k);
    end
    
    % OVERRIDE INITIAL CLUSTER ASSIGNMENT WITH KMEANS w: size = n*D x L
    if flgUseKMeans
        [indices,centers] = kmeans(w',K,'emptyaction','singleton'); 
    end
    
    % w0
    for k=1:K
        mk.comp(k).w0 = centers(k,:)';
    end
    
    % zN
    zN = zeros(L,K);
    for i=1:L
       zN(i,indices(i)) = 1;
    end
    
    % alpha
    for k=1:K
        mk.comp(k).bN = mk.b0;
        mk.comp(k).aN = mk.a0;
        mk.comp(k).alpha = mk.a0/mk.b0; %gamrnd(mk.a0,mk.b0);
    end
   
    SN = zeros(M,M,L);
    for i=1:L
        SN(:,:,i) = (mk.a0/mk.b0)^-1 * eye(M);
    end
    
    % mixture indices and mixing coefficients
    for k = 1:K
        mk.comp(k).pi =  1/L*sum(zN(:,k));
    end

    lambda0_tmp = zeros(1,K);
    lambdal_tmp = zeros(V,K);
    %compute pca regression as initialization
    for k=1:K
        yTrainComponent = struct([]);
        XTrainComponent = struct([]);
        indices = find(zN(:,k) > 0);
        L_k = length(indices);
        for ind = 1:L_k
             yTrainComponent(ind).data = yTrainInit(indices(ind)).data;
             XTrainComponent(ind).data = XTrainInit(indices(ind)).data;
        end
        
        mk.comp(k).h = randn(V,L); %zeros(V,L);
        mk.comp(k).wl = zeros(M,V);
        yTrain = yTrainComponent;
        XTrain = XTrainComponent;
        for i = 1:L_k
            yTrain(i).data = yTrain(i).data - XTrain(i).data* mk.comp(k).w0; 
        end

         %Residual PCA
        for vId = 1:V
              wResidual = zeros(M,L_k);
              for i = 1:L_k
                [wResidual(:,i),~] = lin_reg_estimate_parameter(XTrain(i).data,yTrain(i).data);
              end
              %SVD on the parameters
              [U,S,Vv] = svd(wResidual);

              mk.comp(k).wl(:,vId) = U(:,1)*S(1,1);%mean(wResidual,2);
              mk.comp(k).h(vId,indices) = Vv(:,1);%(pinv(wResidual)*wl(:,vId))';

              for i = 1:L_k
                yTrain(i).data = yTrain(i).data - XTrain(i).data*mk.comp(k).wl(:,vId)*mk.comp(k).h(vId,indices(ind)); 
              end
        end

        
        if use_approx_hypers
            lambda0_tmp(k) = var(mk.comp(k).w0)^-1;
            mk.comp(k).lambda0N = lambda0_tmp(k); 
            
            for v=1:V
                lambdal_tmp(k) = var(mk.comp(k).wl(:,v))^-1;
            end
        else
            mk.comp(k).lambda0N = mk.lambda0;
        end
        
        mk.comp(k).LV = eye(V,V);
        cov_wl = cov(mk.comp(k).wl);
        
        if use_approx_hypers
            for v=1:V
                mk.comp(k).LV(v,v) = 1/cov_wl(v,v);
                mk.comp(k).lambdalN(v) = lambdal_tmp(v,k);
            end
        else
            for v=1:V
                mk.comp(k).LV(v,v) = mk.lambdal;
                mk.comp(k).lambdalN(v) = mk.lambdal;
            end
        end

        c = mk.comp(k);
        if use_approx_hypers
            mk.comp(k).gammaN = c.alpha * (c.wl' * c.wl + M * c.LV) + mk.gamma * eye(V);

            [Usvd,Ssvd,Vsvd] = svd(mk.comp(k).gammaN);
            [Usvd,Ssvd,Vsvd] = svd(c.alpha * (c.wl' * c.wl + M * c.LV) + mk.gamma * eye(V));
            SsvdInv = diag( diag(Ssvd).^-1 );
            mk.comp(k).gammaNInv = Usvd*SsvdInv*Vsvd';
                        
%             mk.comp(k).gammaNInv = cov(mk.comp(k).h')';
        else
            mk.comp(k).gammaN = mk.gamma * eye(V);
            [Usvd,Ssvd,Vsvd] = svd(mk.comp(k).gammaN);
            SsvdInv = diag( diag(Ssvd).^-1 );
            mk.comp(k).gammaNInv = Usvd*SsvdInv*Vsvd';
        end
        
    end
    
      
    
    % lambdal und lambda0
    if use_approx_hypers
        mk.lambda0 = mean(lambda0_tmp(lambda0_tmp~=Inf));
        mk.lambdal = mean(lambdal_tmp(lambdal_tmp~=Inf));        
    end


end

function [mu_w,sigma_w] = lin_reg_estimate_parameter(X_train,y_train, regScalar)

    M = size(X_train,2);
    if nargin < 3
        regScalar = 0;
    end
    if 1==0
        regScalar = 10^-5;
        w = (X_train' * X_train + eye(M)*regScalar) \ X_train' * y_train;
    else
       %use svd to compute the inverse
       [U,S,V] = svds(X_train' * X_train + eye(M)*regScalar, size(X_train,2));
       
       tolSV = 1e-6;
       
       r = find(diag(S)>tolSV,1,'last');%size(S,2);
       s = diag(ones(r,1)./diag(S(1:r,1:r)));
       XtransXInv = V(:,1:r)*s*U(:,1:r)';
       w = XtransXInv * X_train' * y_train;
       
%        keyboard
    end
    
    mu_w = mean(w,2);
    sigma_w = var(w,1,2);   
end